@@ -12,14 +12,11 @@
 
 import abc
 
-import six
-
 from ovsdbapp import api
 from ovsdbapp import constants as const
 
 
-@six.add_metaclass(abc.ABCMeta)
-class API(api.API):
+class API(api.API, metaclass=abc.ABCMeta):
     """An API based off of the ovn-nbctl CLI interface
 
     This API basically mirrors the ovn-nbctl operations with these changes:
@@ -86,9 +83,16 @@ class API(api.API):
         """
 
     @abc.abstractmethod
-    def ls_remove_dns_record(self, switch_uuid, dns_uuid):
+    def ls_remove_dns_record(self, switch_uuid, dns_uuid, if_exists=False):
         """Remove the 'dns_record' from the switch's 'dns_records' list
 
+        :param switch_uuid: The uuid of the switch
+        :type switch_uuid:  string or uuid.UUID
+        :param dns_uuid:    The uuid of the DNS record
+        :type dns_uuid:     string or uuid.UUID
+        :param if_exists:   If True, don't fail if the DNS record
+                            doesn't exist
+        :type if_exists:    boolean
         :returns: :class:`Command` with RowView result
         """
 
@@ -105,7 +109,8 @@ class API(api.API):
         :param match:     The match rule
         :type match:      string
         :param action:    The action to take upon match
-        :type action:     'allow', 'allow-related', 'drop', or 'reject'
+        :type action:     'allow', 'allow-related', 'allow-stateless', 'drop',
+                          or 'reject'
         :param log:       If True, enable packet logging for the ACL
         :type log:        boolean
         :returns:         :class:`Command` with RowView result
@@ -142,22 +147,33 @@ class API(api.API):
 
     @abc.abstractmethod
     def pg_acl_add(self, port_group, direction, priority, match, action,
-                   log=False):
+                   log=False, may_exist=False, severity=None, name=None,
+                   meter=None, **external_ids):
         """Add an ACL to 'port_group'
 
-        :param port_group: The name or uuid of the port group
-        :type port_group:  string or uuid.UUID
-        :param direction:  The traffic direction to match
-        :type direction:   'from-lport' or 'to-lport'
-        :param priority:   The priority field of the ACL
-        :type priority:    int
-        :param match:      The match rule
-        :type match:       string
-        :param action:     The action to take upon match
-        :type action:      'allow', 'allow-related', 'drop', or 'reject'
-        :param log:        If True, enable packet logging for the ACL
-        :type log:         boolean
-        :returns:          :class:`Command` with RowView result
+        :param port_group:   The name or uuid of the port group
+        :type port_group:    string or uuid.UUID
+        :param direction:    The traffic direction to match
+        :type direction:     'from-lport' or 'to-lport'
+        :param priority:     The priority field of the ACL
+        :type priority:      int
+        :param match:        The match rule
+        :type match:         string
+        :param action:       The action to take upon match
+        :type action:        'allow', 'allow-related', 'allow-stateless',
+                             'drop', or 'reject'
+        :param log:          If True, enable packet logging for the ACL
+        :type log:           boolean
+        :param may_exist:    If True, don't fail if the ACL already exists
+        :type may_exist:     boolean
+        :param severity:     Logging at alert, debug, info, notice or warning
+        :type severity:      string
+        :param name:         The name of the ACL
+        :type name:          string
+        :param meter:        The meter name to rate limit logging
+        :type meter:         string
+        :param external_ids: Values to be added as external_id pairs
+        :returns:            :class:`Command` with RowView result
         """
 
     @abc.abstractmethod
@@ -216,7 +232,8 @@ class API(api.API):
         """
 
     @abc.abstractmethod
-    def qos_del(self, switch, direction=None, priority=None, match=None):
+    def qos_del(self, switch, direction=None, priority=None, match=None,
+                if_exists=True):
         """Remove Qos rules from 'switch'
 
         If only switch is supplied, all the QoS rules from the logical switch
@@ -233,6 +250,8 @@ class API(api.API):
         :type priority:   int
         :param match:     The match rule
         :type match:      string
+        :param if_exists: Do not fail if the Logical_Switch row does not exist
+        :type if_exists:  bool
         :returns:         :class:`Command` with no result
         """
 
@@ -245,6 +264,19 @@ class API(api.API):
         :returns:      :class:`Command` with RowView list result
         """
 
+    @abc.abstractmethod
+    def qos_del_ext_ids(self, lswitch_name, external_ids, if_exists=True):
+        """Delete all QoS rules related to a floating IP.
+
+        :param lswitch_name: The unique name of the logical switch
+        :type lswitch_name: string
+        :param external_ids: Pairs of key/value to find in the "external_ids"
+        :type external_ids: dict
+        :param if_exists: Do not fail if the Logical_Switch row does not exist
+        :type if_exists: bool
+        :returns: :class:`Command` with no result
+        """
+
     @abc.abstractmethod
     def lsp_add(self, switch, port, parent_name=None, tag=None,
                 may_exist=False, **columns):
@@ -535,6 +567,26 @@ class API(api.API):
         :returns:
         """
 
+    @abc.abstractmethod
+    def lrp_set_options(self, port, **options):
+        """Set options related to the type of 'port'
+
+        :param port:    The name or uuid of the port
+        :type port:     string or uuid.UUID
+        :param options: keys and values for the port 'options' dict
+        :type options:  key: string, value: string
+        :returns:       :class:`Command` with no result
+        """
+
+    @abc.abstractmethod
+    def lrp_get_options(self, port):
+        """Get the type-specific options for 'port'
+
+        :param port: The name or uuid of the port
+        :type port:  string or uuid.UUID
+        :returns:    :class:`Command` with dict result
+        """
+
     @abc.abstractmethod
     def lr_route_add(self, router, prefix, nexthop, port=None,
                      policy='dst-ip', may_exist=False):
@@ -638,7 +690,8 @@ class API(api.API):
         """
 
     @abc.abstractmethod
-    def lb_add(self, vip, ips, protocol=const.PROTO_TCP, may_exist=False):
+    def lb_add(self, lb, vip, ips, protocol=const.PROTO_TCP, may_exist=False,
+               **columns):
         """Create a load-balancer or add a VIP to an existing load balancer
 
         :param lb:        The name or uuid of the load-balancer
@@ -652,6 +705,8 @@ class API(api.API):
         :param may_exist: If True, don't fail if a LB w/ `vip` exists, and
                           instead, replace the vips on the LB
         :type may_exist:  boolean
+        :param columns:   Additional columns to directly set on the load
+                          balancer
         :returns:        :class:`Command` with RowView result
         """
 
@@ -659,11 +714,15 @@ class API(api.API):
     def lb_del(self, lb, vip=None, if_exists=False):
         """Remove a load balancer or just the VIP from a load balancer
 
+        If all vips of the load balancer are removed, the load balancer
+        instance will be removed as well. If vip parameter is 'None'
+        this will cause all vips (and load balancer) to be removed.
+
         :param lb:        The name or uuid of a load balancer
         :type lb:         string or uuid.UUID
-        :param vip:       The VIP on the load balancer to match
+        :param vip:       The VIP to match. If None, match all vips
         :type:            string
-        :param if_exists: If True, don't fail if the port doesn't exist
+        :param if_exists: If True, don't fail if the lb/vip doesn't exist
         :type if_exists:  boolean
         """
 
@@ -852,12 +911,15 @@ class API(api.API):
         """
 
     @abc.abstractmethod
-    def dns_remove_record(self, uuid, hostname):
+    def dns_remove_record(self, uuid, hostname, if_exists=False):
         """Remove the 'hostname' from the 'records' field of the DNS row
 
         :param uuid: The uuid of the DNS row to set the records with
         :type uuid:  string or uuid.UUID
         :param hostname: hostname as the key to the record dict
+        :param if_exists:   If True, don't fail if the DNS record
+                            doesn't exist
+        :type if_exists:    boolean
         :returns:    :class:`Command` with no result
         """
 
@@ -938,3 +1000,125 @@ class API(api.API):
         :type pg_id:  string or uuid.UUID
         :returns:     :class:`Command` with RowView result
         """
+
+    @abc.abstractmethod
+    def ha_chassis_group_add(self, name, may_exist=False, **columns):
+        """Create a HA Chassis Group
+
+        :param name:        The name of the ha chassis group
+        :type name:         string
+        :param may_exist:   If True, don't fail if the ha chassis group
+                            already exists
+        :type may_exist:    bool
+        :param columns:     Additional columns to directly set on the ha
+                            chassis group (e.g external_ids)
+        :type columns:      dictionary
+        :returns:           :class:`Command` with RowView result
+        """
+
+    @abc.abstractmethod
+    def ha_chassis_group_del(self, name, if_exists=False):
+        """Delete a HA Chassis Group
+
+        :param name:        The name of the ha chassis group
+        :type name:         string
+        :param if_exists:   If True, don't fail if the ha chassis group
+                            doesn't exist
+        :type if_exists:    boolean
+        :returns:           :class:`Command` with no result
+        """
+
+    @abc.abstractmethod
+    def ha_chassis_group_get(self, name):
+        """Get HA Chassis Group
+
+        :param name: The name or uuid of the ha chassis group
+        :type name:  string or uuid.UUID
+        :returns:    :class:`Command` with RowView result
+        """
+
+    @abc.abstractmethod
+    def ha_chassis_group_add_chassis(self, hcg_id, chassis, priority,
+                                     **columns):
+        """Add a HA Chassis to a HA Chassis Group
+
+        :param hcg_id:   The name or uuid of the ha chassis group
+        :type hcg_id:    string or uuid.UUID
+        :param chassis:  The name of the ha chassis
+        :type chassis:   string
+        :param priority: The priority of the ha chassis
+        :type priority:  int
+        :param columns:  Additional columns to directly set on the ha
+                         chassis (e.g external_ids)
+        :type columns:   dictionary
+        :returns:        :class:`Command` with RowView result
+        """
+
+    @abc.abstractmethod
+    def ha_chassis_group_del_chassis(self, hcg_id, chassis, if_exists=False):
+        """Delete a HA Chassis from a HA Chassis Group
+
+        :param hcg_id:     The name or uuid of the ha chassis group
+        :type hcg_id:      string or uuid.UUID
+        :param chassis:    The name of the ha chassis
+        :type chassis:     string
+        :param if_exists:  If True, don't fail if the ha chassis
+                           doesn't exist
+        :type if_exists:   boolean
+        :returns:          :class:`Command` with no result
+        """
+
+    @abc.abstractmethod
+    def meter_add(self, name, unit, rate=1, fair=False, burst_size=0,
+                  action=None, may_exist=False, **columns):
+        """Create a Meter
+
+        :param name:        The name of the meter
+        :type name:         string
+        :param unit:        The unit of the meter (e.g kbps, pktps)
+        :type unit:         string
+        :param rate:        The rate of the meter
+        :type rate:         int
+        :param fair:        Specify whether meter rate limits its references
+                            individually (True) or as a shared pool (False)
+        :type fair:         boolean
+        :param burst_size:  The maximum burst allowed for the band in kilobits
+                            or packets, depending on the unit used
+        :type burst_size:   int
+        :param action:      The action of the meter. 'None' can be used as an
+                            alternate for the 'drop' value.
+        :type action:       string
+        :param may_exist:   If True, don't fail if the meter already exists
+        :type may_exist:    boolean
+        :param columns:     Additional columns to directly set on the meter
+                            (e.g external_ids)
+        :type columns:      dictionary
+        :returns:           :class:`Command` with RowView result
+        """
+
+    @abc.abstractmethod
+    def meter_del(self, meter, if_exists=False):
+        """Delete a meter
+
+        :param meter:       The name or uuid of the meter
+        :type meter:        string or uuid.UUID
+        :param if_exists:   If True, don't fail if the meter doesn't exist
+        :type if_exists:    boolean
+        :returns:           :class:`Command` with no result
+        """
+
+    @abc.abstractmethod
+    def meter_list(self):
+        """Get all meters
+
+        :returns:      :class:`Command` with RowView list result
+        """
+
+    @abc.abstractmethod
+    def meter_get(self, meter):
+        """Get the meter
+
+        :param meter:  The name or uuid of the meter
+        :type meter:   string or uuid.UUID
+        :returns:      :class:`Command` with RowView result
+        """
